// butt - broadcast using this tool
//
// Copyright 2007 by Daniel Noethen.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <lame/lame.h>
#include <signal.h>
#include <limits.h>
#include <pthread.h>

#ifdef _WIN32
 #include <time.h>
#endif

#include "cfg.h"
#include "butt.h"
#include "port_audio.h"
#include "lame_encode.h"
#include "shoutcast.h"
#include "parseconfig.h"

#include "config.h"

bool record;
bool recording;
bool connected;
bool streaming;
bool disconnect;
bool try_connect;
bool song_timeout_running;

int stream_socket;
unsigned int bytes_sent;
unsigned int bytes_written;

sec_timer rec_timer;
sec_timer stream_timer;
sec_timer xrun_timer;

int main()
{

    char *p;
    char lcd_buf[33];
    char info_buf[256];

#ifndef _WIN32
    signal(SIGPIPE, SIG_IGN); //ignore the SIGPIPE signal.
        //(in case the server closes the connection unexpected)
#endif

    snprintf(info_buf, sizeof(info_buf),
            "starting %s\nwritten by Daniel Noethen\n", VERSION);
    print_info(info_buf, 0);

    SHOW_GUI();

#ifdef _WIN32
    //If there is no "%APPDATA% we are probably in none-NT Windows
    //So we save the config file to the programm directory
    if((p = getenv("APPDATA")) == NULL)
    {
        cfg_path = (char*)malloc(strlen(CONFIG_FILE)+1);
        strcpy(cfg_path, CONFIG_FILE);
    }
    else
    {
        cfg_path = (char*)malloc(PATH_MAX+strlen(CONFIG_FILE)+1);
        snprintf(cfg_path, PATH_MAX+strlen(CONFIG_FILE), "%s\\%s", p, CONFIG_FILE);  
    }
#else //UNIX
    if((p = getenv("HOME")) == NULL)
    {
        ALERT("No home-directory found");
        return 1;
    }
    cfg_path = (char*)malloc(PATH_MAX+strlen(CONFIG_FILE)+1);
    snprintf(cfg_path, PATH_MAX+strlen(CONFIG_FILE), "%s/%s", p, CONFIG_FILE);  

#endif

    if(!snd_init())
        print_info("PortAudio init succeeded", 0);
    else
    {
        ALERT("PortAudio init failed");
        return 1;
    }

    if(cfg_set_values())        //read configfile and set the config struct
    {
        if(cfg_create_default())
        {
            ALERT("Could not create config file");
            return 1;
        }
        sprintf(info_buf, "butt created a default config file:\n(%s)\n",
                cfg_path );

        print_info(info_buf, 0);
        cfg_set_values();
    }

    if(!lame_enc_init())
        print_info("lame init succeeded", 0);
    else
    {
        ALERT("LAME init failed");
        return 1;
    }
    vorbis_enc_init();

    print_info("=========================\n", 0);

    snd_open_stream();

    strcpy(lcd_buf, "info: idle");
    PRINT_LCD(lcd_buf, strlen(lcd_buf), 0, 1);

    GUI_LOOP();

    return 0;
}
