// icecast functions for butt
//
// Copyright 2007 by Daniel Noethen.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "cfg.h"
#include "butt.h"
#include "timer.h"
#include "parseconfig.h"

#include "config.h"

shout_t *shout;

int ic_update_song()
{
    int rc;

    shout_metadata_t* metadata;
    metadata = shout_metadata_new();

    shout_metadata_add (metadata, "song", cfg.main.song);
    if(shout_set_metadata (shout, metadata) != SHOUTERR_SUCCESS)
        rc = 1; //updating song failed
    else
        rc = 0; //updating song suceeded

    shout_metadata_free(metadata);

    return rc;
}

int ic_init()
{
    shout_init();

    if (!(shout = shout_new()))
    {
        print_info("Could not allocate shout_t", 1);
        return 1;
    }

    



    return 0;
}

int ic_connect()
{
    int srv, rv;

    srv = cfg.selected_srv;

     //TODO: add error handling

    shout_set_host(shout, cfg.srv[srv]->addr);
    shout_set_protocol(shout, SHOUT_PROTOCOL_HTTP);
    shout_set_port(shout, cfg.srv[srv]->port);
    shout_set_password(shout, cfg.srv[srv]->pwd);
    shout_set_mount(shout, cfg.srv[srv]->mount);
    //the username for sources is allways "source"
    shout_set_user(shout, "source");
    //shout_set_name(shout,cfg.icy[cfg.selected_icy]->name);
    if(cfg.main.num_of_icy >0)
    {
        shout_set_genre(shout, cfg.icy[cfg.selected_icy]->genre);
        shout_set_description(shout, cfg.icy[cfg.selected_icy]->desc);
        shout_set_url(shout, cfg.icy[cfg.selected_icy]->url);
        shout_set_public(shout,  atoi(cfg.icy[cfg.selected_icy]->pub));
    }

    shout_set_agent(shout, VERSION);

    if(!strcmp(cfg.audio.codec, "ogg"))
    {
#if HAVE_LIBVORBIS
        shout_set_format(shout, SHOUT_FORMAT_OGG);
        vorbis_enc_init(); //TODO: check why we need to reinit vorbis before every connect
#endif
    }
    else
        shout_set_format(shout, SHOUT_FORMAT_MP3);


    if (shout_set_nonblocking(shout, 1) != SHOUTERR_SUCCESS) 
	  printf("Error setting non-blocking mode: %s\n", shout_get_error(shout));


    if((rv = shout_open(shout)) == SHOUTERR_SUCCESS)
        rv = SHOUTERR_CONNECTED;

    while (rv == SHOUTERR_BUSY)
    {
        rv = shout_get_connected(shout);
        usleep(1000);
    }

    if(rv == SHOUTERR_NOLOGIN)
    {
        print_info("\nconnect: invalid password!\n", 1);
        return 2; //stops reconnecting if pw is wrong
    }

    connected = 1;

    timer_init(&stream_time);       //starts the "online" timer

    return 0;
}

void ic_disconnect()
{
    shout_sync(shout);
    shout_close(shout);
#if HAVE_LIBVORBIS
    vorbis_enc_close();
#endif
   // shout_free(shout);
   // shout_shutdown();
}

int ic_send(char *buf, int buf_len)
{
    if(shout_get_connected(shout) == SHOUTERR_UNCONNECTED)
        printf("fuuuuuck\n");

    if(shout_send(shout, (unsigned char*)buf, buf_len) !=
            SHOUTERR_SUCCESS)
    {
        printf("shout error\n");
        return -1;
    }
    //shout_send doesnt return the length of the written data
    //let's assume that it really writes it all to the net
    return buf_len;
}

