// portaudio functions for butt
//
// Copyright 2007 by Daniel Noethen.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <string.h>
#include <pthread.h>

#ifdef _WIN32
 #include <windows.h>
#endif

#include "butt.h"
#include "cfg.h"
#include "port_audio.h"
#include "parseconfig.h"
#include "lame_encode.h"
#include "shoutcast.h"
#include "icecast.h"
#include "strfuncs.h"

#include "config.h"

#define PA_FRAMES 2048
#define FRAMES_MUL 2
#define BUF_FRAMES PA_FRAMES*FRAMES_MUL

char* encode_buf;
short *pa_pcm_buf;
int buf_index;
int buf_size;
int buf_pos;
int framepacket_size;

int try_to_connect;
bool pa_new_frames;
bool snd_xrun;
bool reconnect;
int pa_callback_counter;
int encode_bytes_read;

pthread_t rec_thread;
pthread_t stream_thread;
pthread_mutex_t stream_mut, rec_mut;
pthread_cond_t  stream_cond, rec_cond;

PaStream *stream;

int snd_init()
{
    char info_buf[256];

    PaError p_err;
    if((p_err = Pa_Initialize()) != paNoError)
    {
        snprintf(info_buf, sizeof(info_buf), "PortAudio init failed:\n%s\n",
                    Pa_GetErrorText(p_err));

        ALERT(info_buf);
        return 1;
    }

    reconnect = 0;
    buf_index = 0;
    return 0;
}

void snd_reinit()
{
    snd_close();
    snd_init();
    snd_open_stream();
}

int snd_open_stream()
{
    int samplerate;
    char info_buf[256];

    PaDeviceIndex pa_dev_id;
    PaStreamParameters pa_params;
    PaError pa_err;
    const PaDeviceInfo *pa_dev_info;

    if(cfg.audio.dev_count == 0)
    {
        print_info("ERROR: no sound device with input channels found", 1);
        return 1;
    }

    pa_callback_counter = 0;

    buf_size = BUF_FRAMES * 2;
    framepacket_size = PA_FRAMES * cfg.audio.channel;

    pa_pcm_buf = (short*)malloc(buf_size * sizeof(short));
    encode_buf = (char*)malloc(buf_size * sizeof(char));

    samplerate = cfg.audio.samplerate;

    pa_dev_id = cfg.audio.pcm_list[cfg.audio.dev_num]->dev_id;

    pa_dev_info = Pa_GetDeviceInfo(pa_dev_id);
    if(pa_dev_info == NULL)
    {
        snprintf(info_buf, 127, "Error getting device Info (%d)", pa_dev_id);
        print_info(info_buf, 1);
        return 1;
    }

    pa_params.device = pa_dev_id;
    pa_params.channelCount = cfg.audio.channel;
    pa_params.sampleFormat = paInt16;
    pa_params.suggestedLatency = pa_dev_info->defaultHighInputLatency;
    pa_params.hostApiSpecificStreamInfo = NULL;

    pa_err = Pa_IsFormatSupported(&pa_params, NULL, samplerate);
    if(pa_err != paFormatIsSupported)
    {
        if(pa_err == paInvalidSampleRate)
        {
            snprintf(info_buf, sizeof(info_buf),
                    "Samplerate not supported: %dHz\n"
                    "Using default samplerate: %dHz",
                    samplerate, (int)pa_dev_info->defaultSampleRate);
            print_info(info_buf, 1);

            if(Pa_IsFormatSupported(&pa_params, NULL,
               pa_dev_info->defaultSampleRate) != paFormatIsSupported)
            {
                print_info("FAILED", 1);
                return 1;
            }
            else
            {
                samplerate = (int)pa_dev_info->defaultSampleRate;
                cfg.audio.samplerate = samplerate;
                cfg.rec.samplerate = samplerate;
                update_samplerates();
            }
        }
        else
        {
            snprintf(info_buf, sizeof(info_buf), "PA: Format not supported: %s\n",
                    Pa_GetErrorText(pa_err));
            print_info(info_buf, 1);
            return 1;
        }
    }

    pa_err = Pa_OpenStream(&stream, &pa_params, NULL,
                            samplerate, PA_FRAMES,
                            paNoFlag, snd_callback, NULL);

    if(pa_err != paNoError)
    {
        printf("error opening sound device: \n%s\n", Pa_GetErrorText(pa_err));
        return 1;
    }

    
    Pa_StartStream(stream);
    return 0;
}

void snd_start_stream()
{   
    pthread_mutex_init(&stream_mut, NULL);
    pthread_cond_init (&stream_cond, NULL);
    
    bytes_sent = 0;
    streaming = 1;
    pthread_create(&stream_thread, NULL, snd_stream_thread, NULL);
}

void snd_stop_stream()
{
    connected = 0;
    streaming = 0;

    pthread_cond_signal(&stream_cond);

    pthread_mutex_destroy(&stream_mut);
    pthread_cond_destroy(&stream_cond);


    print_info("user disconnected\n", 0);
}

void *snd_stream_thread(void *data)
{
    int (*xc_send)(char *buf, int buf_len) = NULL;
    int sent;
    short *local_audio_buf = (short*)malloc(buf_size*sizeof(short));

    encode_bytes_read = 0;

    if(cfg.srv[cfg.selected_srv]->type == SHOUTCAST)
            xc_send = &sc_send;
    if(cfg.srv[cfg.selected_srv]->type == ICECAST)
        xc_send = &ic_send;

    while(connected)
    {
        pthread_cond_wait(&stream_cond, &stream_mut);
        if(!connected)
            break;

        pa_callback_counter = 0;

        memcpy(local_audio_buf, pa_pcm_buf, buf_size*sizeof(short));

        if(!strcmp(cfg.audio.codec, "mp3"))
            encode_bytes_read = lame_enc_buf(local_audio_buf, encode_buf,
                                    BUF_FRAMES, gfp_stream, buf_size);
#if HAVE_LIBVORBIS
        else
            encode_bytes_read = vorbis_enc_buf(local_audio_buf,
                                               encode_buf,
                                               BUF_FRAMES);
#endif
        if((sent = xc_send(encode_buf, encode_bytes_read)) == -1)
            connected = 0; //disconnected
        else
            bytes_sent += encode_bytes_read;
    }

    free(local_audio_buf);
    return NULL;
}

void snd_start_rec()
{
    pthread_mutex_init(&rec_mut, NULL);
    pthread_cond_init (&rec_cond, NULL);

    bytes_written = 0;
    recording = 1;

    pthread_create(&rec_thread, NULL, snd_rec_thread, NULL);

    print_info("recording to:", 0);
    print_info(cfg.rec.path, 0);
}

void snd_stop_rec()
{
    record = 0;
    recording = 0;

    pthread_cond_signal(&rec_cond);

    pthread_mutex_destroy(&rec_mut);
    pthread_cond_destroy(&rec_cond);


    print_info("recording stopped", 0);
}

//The recording stuff runs in its own thread
//this prevents dropouts in the recording, in case the
//bandwidth is smaller than the selected streaming bitrate
void* snd_rec_thread(void *data)
{
    int   enc_bytes_read;
    char  *enc_buf = (char*)malloc(buf_size * sizeof(char));
    short *audio_buf = (short*)malloc(buf_size * sizeof(short));

    while(record)
    {
        pthread_cond_wait(&rec_cond, &rec_mut);
        memcpy(audio_buf, pa_pcm_buf, buf_size*sizeof(short));

        if(!strcmp(cfg.rec.codec, "mp3"))
        {
            enc_bytes_read = lame_enc_buf(audio_buf, enc_buf,
                                BUF_FRAMES, gfp_record, buf_size);
            bytes_written += fwrite(enc_buf, 1, enc_bytes_read, cfg.rec.fd);
        }
        else if (!strcmp(cfg.rec.codec, "pcm"))
            bytes_written += fwrite(audio_buf, sizeof(short), buf_size, cfg.rec.fd);
    }


    fclose(cfg.rec.fd);
    free(enc_buf);
    free(audio_buf);
    return NULL;
}

//this function is called by PortAudio when new audio data arrived
int snd_callback(const void *input,
                 void *output,
                 unsigned long frameCount,
                 const PaStreamCallbackTimeInfo* timeInfo,
                 PaStreamCallbackFlags statusFlags,
                 void *userData)
{
    //If this function gets more often called than FRAMES_MUL,
    //then the snd_send_buf function wasn't fast enough to send the data to the server
    //this results in a buffer overrun
    if((pa_callback_counter++ >= FRAMES_MUL) ) 
       snd_xrun = 1;
    else
       snd_xrun = 0;

    buf_pos = buf_index*framepacket_size;
    memcpy(pa_pcm_buf+buf_pos, input, framepacket_size*sizeof(short));

    pa_new_frames = 1;     //tell vu_update() that there is new audio data

    buf_index++;
    if(buf_index == FRAMES_MUL)
    {
        if(streaming)
            pthread_cond_signal(&stream_cond);
        if(recording)
            pthread_cond_signal(&rec_cond);

        buf_index = 0;
    }

    return 0;
}

void snd_update_vu()
{
    short *p;
    int i;
    int lpeak = 0;
    int rpeak = 0;

    p = pa_pcm_buf+buf_pos;
    for(i = 0; i < framepacket_size; i += cfg.audio.channel)
    {
        if(p[i] > lpeak)
            lpeak = p[i];
        if(p[i+(cfg.audio.channel-1)] > rpeak)
            rpeak = p[i+(cfg.audio.channel-1)];
    }

    vu_meter(lpeak, rpeak);

    pa_new_frames = 0;
}

/*
int *snd_get_samplerates(int *sr_count)
{

    int sr[] = { 11025, 22050, 44100, 48000 };
    int *sr_list;

    int count = 0;
    PaDeviceIndex pa_dev_id;
    PaStreamParameters pa_params;
    const PaDeviceInfo *pa_dev_info;

    sr_list = (int*)malloc(4*sizeof(int));

    //we need to close the snd_stream to get answers from the PA backend
    snd_close();
    snd_init();

    if(cfg.audio.pcm_list[cfg.audio.dev_num]->dev_id == -1)
    {
        pa_dev_id = Pa_GetDefaultInputDevice();
        if(pa_dev_id == paNoDevice)
            pa_dev_id = 0;
    }
    else
        pa_dev_id = cfg.audio.pcm_list[cfg.audio.dev_num]->dev_id;

    pa_dev_info = Pa_GetDeviceInfo(pa_dev_id);
    pa_params.device = pa_dev_id;
    pa_params.channelCount = cfg.audio.channel;
    pa_params.sampleFormat = paInt16;
    pa_params.suggestedLatency = pa_dev_info->defaultHighInputLatency;
    pa_params.hostApiSpecificStreamInfo = NULL;

    for(int i = 0; i < 4; i++)
    {
        if(Pa_IsFormatSupported(&pa_params, NULL, sr[i]) != paInvalidSampleRate)
        {
            sr_list[count] = sr[i];
            count++;
        }
    }

    snd_open_stream();

    *sr_count = count;
    return sr_list;
}
*/

snd_dev_t **snd_get_devices(int *dev_count)
{
    int devcount, sr_count, i, dev_num = 0;
    bool sr_supported = 0;
    const PaDeviceInfo *p_di;
    char info_buf[256];
    PaStreamParameters pa_params;

    int sr[] = { 8000, 11025, 12000, 16000, 22050, 24000, 32000, 44100, 48000 };

    snd_dev_t **dev_list;

    dev_list = (snd_dev_t**)malloc(100*sizeof(snd_dev_t*));

    for(i = 0; i < 100; i++)
        dev_list[i] = (snd_dev_t*)malloc(sizeof(snd_dev_t));

    dev_list[dev_num]->name = (char*) malloc(strlen("Default PCM device (default)")+1);
    strcpy(dev_list[dev_num]->name, "Default PCM device (default)");
    dev_list[dev_num]->dev_id = Pa_GetDefaultInputDevice();
    dev_num++;


    devcount = Pa_GetDeviceCount();
    if(devcount < 0)
    {
        snprintf(info_buf, sizeof(info_buf), "PaError: %s", Pa_GetErrorText(devcount));
        print_info(info_buf, 1);
    }

    for(i = 0; i < devcount && i < 100; i++)
    {
        sr_count = 0;
        p_di = Pa_GetDeviceInfo(i);
        if(p_di == NULL)
        {
            snprintf(info_buf, sizeof(info_buf), "Error getting device Info_ (%d)", i);
            print_info(info_buf, 1);
            continue;
        }


        //Save only devices which have input Channels
        if(p_di->maxInputChannels <= 0)
            continue;

        pa_params.device = i;
        pa_params.channelCount = cfg.audio.channel;
        pa_params.sampleFormat = paInt16;
        pa_params.suggestedLatency = p_di->defaultHighInputLatency;
        pa_params.hostApiSpecificStreamInfo = NULL;

        //add the supported samplerates to the device structure
        for(int j = 0; j < 9; j++)
        {
            if(Pa_IsFormatSupported(&pa_params, NULL, sr[j]) != paInvalidSampleRate)
            {
                dev_list[dev_num]->sr_list[sr_count] = sr[j];
                sr_count++;
                sr_supported = 1;
            }
        }
        //Go to the next device if this one doesn't support one of our samplerates
        if(!sr_supported)
            continue;

        //Mark the end of the samplerate list for this device
        dev_list[dev_num]->sr_list[sr_count] = 0;

        dev_list[dev_num]->name = (char*) malloc(strlen(p_di->name)+1);
        strcpy(dev_list[dev_num]->name, p_di->name);
        dev_list[dev_num]->dev_id = i;

        //copy the sr_list from the device where the
        //virtual default device points to
        if(dev_list[0]->dev_id == dev_list[dev_num]->dev_id)
            memcpy(dev_list[0]->sr_list, dev_list[dev_num]->sr_list,
                    sizeof(dev_list[dev_num]->sr_list));


        //We need to escape every '/' in the device name
        //otherwise FLTK will add a submenu for every '/' in the dev list
        strrpl(&dev_list[dev_num]->name, "/", "\\/");

        dev_num++;
    }//for(i = 0; i < devcount && i < 100; i++)


    if(dev_num == 1)
        *dev_count = 0;
    else
        *dev_count = dev_num;

    return dev_list;
}

void snd_close()
{
    Pa_StopStream(stream);
    Pa_CloseStream(stream);
    Pa_Terminate();

    free(pa_pcm_buf);
    free(encode_buf);
}

