// vorbis encoding functions for butt
//
// Copyright 2007 by Daniel Noethen.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//

#include <string.h>

#include "vorbis_encode.h"
#include "cfg.h"

ogg_stream_state os; /* take physical pages, weld into a logical
                      stream of packets */
ogg_page         og; /* one Ogg bitstream page.  Vorbis packets are inside */
ogg_packet       op; /* one raw packet of data for decode */

vorbis_info      vi;
vorbis_comment   vc; /* struct that stores all the user comments */

vorbis_block     vb;

vorbis_dsp_state vd; /* central working state for the packet->PCM decoder */
int vorbis_enc_init()
{
    int ret;
    ogg_packet header;
    ogg_packet header_comm;
    ogg_packet header_code;

    //TODO: add error handling
    vorbis_info_init(&vi);

    ret = vorbis_encode_init(&vi,
                  cfg.audio.channel,
                  cfg.audio.samplerate,
                  cfg.audio.bitrate*1000,
                  cfg.audio.bitrate*1000,
                  cfg.audio.bitrate*1000);
    if(ret)
        return ret;

    vorbis_comment_init(&vc);
    vorbis_comment_add_tag(&vc,"ENCODER", VERSION);

    vorbis_analysis_init(&vd,&vi);
    vorbis_block_init(&vd,&vb);

    ogg_stream_init(&os,0);

    vorbis_analysis_headerout(&vd,&vc,&header,&header_comm,&header_code);
    ogg_stream_packetin(&os,&header);
    ogg_stream_packetin(&os,&header_comm);
    ogg_stream_packetin(&os,&header_code);

    return 0;
}
int vorbis_enc_reinit()
{
     vorbis_enc_close();
     return vorbis_enc_init();
}

int vorbis_enc_buf(short *pcm_buf, char *enc_buf, int size)
{
    int i, result;
    int eos = 0;
    int w = 0;
    float **vorbis_buf;

    if(size == 0)
        return 0;


    /* This ensures the actual
     * audio data will start on a new page, as per spec
     */
    while(!eos)
    {
        result = ogg_stream_flush(&os,&og);
        if(result == 0)
            break;
        memcpy(enc_buf+w, og.header, og.header_len);
        w += og.header_len;
        memcpy(enc_buf+w, og.body, og.body_len);
        w += og.body_len;
    }

    vorbis_buf = vorbis_analysis_buffer(&vd, size);

    //deinterlace audio data and convert it from short to float
    if(cfg.audio.channel == 1)
         for(i = 0 ; i < size ; i++)
             vorbis_buf[0][i] = pcm_buf[i]/32768.f;
    else
    {
        for(i = 0 ; i < size; i++)
        {
            vorbis_buf[0][i] = pcm_buf[i*2]/32768.f;
            vorbis_buf[1][i] = pcm_buf[i*2+1]/32768.f;
        }
    }

    vorbis_analysis_wrote( &vd, i);

    while(vorbis_analysis_blockout(&vd,&vb) == 1)
    {
        vorbis_analysis(&vb,&op);
        vorbis_bitrate_addblock(&vb);

        while(vorbis_bitrate_flushpacket(&vd,&op))
        {
            /* weld the packet into the bitstream */
            ogg_stream_packetin(&os,&op);

            /* write out pages (if any) */
            while(!eos)
            {
                result = ogg_stream_pageout(&os, &og);
                if(result == 0) 
                    break;
                memcpy(enc_buf+w, og.header, og.header_len);
                w += og.header_len;
                memcpy(enc_buf+w, og.body, og.body_len);
                w += og.body_len;
                if(ogg_page_eos(&og))eos=1;
            }
        }
    }


    return w;
}

void vorbis_enc_close()
{
    ogg_stream_clear(&os);
    vorbis_block_clear(&vb);
    vorbis_dsp_clear(&vd);
    vorbis_comment_clear(&vc);
    vorbis_info_clear(&vi);
}
